import org.antlr.runtime.*;
import org.antlr.runtime.tree.*;

class DeltaInterpreter {
	SymbolTable symtab;
	InternalFunctions functions;


	SymbolTable getSymbolTable() {
		return symtab;
	}
	
	DeltaInterpreter(SymbolTable _symtab,
			 InternalFunctions _functions) {
		symtab = _symtab;
		functions = _functions;
	}

	ValueObject evalExpr(CommonTree expr) throws InterpreterException{

		CommonTree left = (CommonTree)expr.getChild(0);
		CommonTree  right = null;
		if (expr.getChildCount()==2) {
			right = (CommonTree)expr.getChild(1);
		}
		if (expr.getType()==DeltaLexer.INT) {
			return new  ValueObject(Integer.parseInt(expr.getText()));
		}
		if (expr.getType()==DeltaLexer.FLOATING_POINT) {
			return new  ValueObject(Double.parseDouble(expr.getText()));
		}
		if (expr.getType()==DeltaLexer.STRING) {
			String s = expr.getText();
			s = s.replaceAll("^\"", "");
			s = s.replaceAll("\"$", "");
			return new  ValueObject(s);
		}
		if (expr.getType()==DeltaLexer.ID) {
			SymbolTable.Symbol var = symtab.get(expr.getText());
			return var.getValue();
		}
		if (expr.getType()==DeltaLexer.FUNCTION_CALL) {
			return evalFunctionCall(expr);
		}

		String operator = expr.getText();

		String compare_operators[] = {"<", ">", "<=", ">=", "==", "!="};

		for (String op: compare_operators) {
			if (op.equals(operator)) {
				ValueObject vleft = evalExpr(left);
				ValueObject vright = evalExpr(right);
				return vleft.compare(op, vright);
			}
		}

		if (operator.equals("+")) {
			if (expr.getChildCount()==1) {
				return evalExpr(left);
			} else {
				return evalExpr(left).add(evalExpr(right));
			}
		}
		if (operator.equals("-")) {
			if (expr.getChildCount()==1) {
				return evalExpr(left).neg();
			} else {
				return evalExpr(left).sub(evalExpr(right));
			}
		}
		if (operator.equals("*")) {
			return evalExpr(left).mul(evalExpr(right));
		}
		if (operator.equals("%")) {
			return evalExpr(left).mod(evalExpr(right));
		}
		if (operator.equals("/")) {
			return evalExpr(left).div(evalExpr(right));
		}
		return new ValueObject(); /*fallback, should not happen*/
	}

	void evalAssignment(CommonTree astmt) throws InterpreterException {
		CommonTree var = (CommonTree)astmt.getChild(0);
		CommonTree expr = (CommonTree)astmt.getChild(1);
		ValueObject res = evalExpr(expr);
		symtab.setValue(var.getText(), res);
	}

	void evalIf(CommonTree astmt) throws InterpreterException {
		CommonTree expr = (CommonTree)astmt.getChild(0);
		CommonTree ifpart = (CommonTree)astmt.getChild(1);
		ValueObject res = evalExpr(expr);
		if (res.getIntValue()!=0) {
			evalStatementList(ifpart);
			return;
		}
		if (astmt.getChildCount()==3) { /*else part*/
			CommonTree elsepart = (CommonTree)astmt.getChild(2);
			evalStatementList(elsepart);
		}
	}

	void evalWhile(CommonTree astmt) throws InterpreterException {
		CommonTree expr = (CommonTree)astmt.getChild(0);
		CommonTree whileblock = (CommonTree)astmt.getChild(1);
		while (evalExpr(expr).getIntValue()!=0)
			evalStatementList(whileblock);		
	}

	ValueObject evalFunctionCall(CommonTree expr) throws InterpreterException {
		return functions.evaluate(this, expr);
	}

	void evalStatement(CommonTree expr) throws InterpreterException{
		switch (expr.getType()) {
		case DeltaLexer.DECLARATION:
			/*handled by type checker*/
			break;
		case DeltaLexer.FUNCTION_CALL:
			evalFunctionCall(expr);
			break;
		case DeltaLexer.ASSIGNMENT:
			evalAssignment(expr);
			break;
		case DeltaLexer.IF_STATEMENT:
			evalIf(expr);
			break;
		case DeltaLexer.WHILE_STATEMENT:
			evalWhile(expr);
			break;
		}
	}

	void evalStatementList(CommonTree exprlist) throws InterpreterException {
		for (Object e: exprlist.getChildren())
			evalStatement((CommonTree)e);
	}

	void eval(CommonTree root) throws InterpreterException {
		evalStatementList(root);
	}
}
