import java.util.Vector;
import java.util.Scanner;
import org.antlr.runtime.*;
import org.antlr.runtime.tree.*;

class DeltaTypeChecker {

	SymbolTable symtab;

	void checkVar(String var) throws InterpreterException {
		if (!symtab.defined(var)) {
			throw new InterpreterException("Variable '"+var+"' not declared");
		}
	}

	DataType evalFunctionCall(CommonTree expr) throws InterpreterException {
		CommonTree functionname = (CommonTree)expr.getChild(0);
		CommonTree parameters = (CommonTree)expr.getChild(1);
		SymbolTable.Symbol f = symtab.get(functionname.getText());
		/*check the type on parameters*/
		for (Object param:parameters.getChildren()) {
			evalExpr((CommonTree)param);
		}
		return f.getDataType();
	}

	DataType evalExpr(CommonTree expr) throws InterpreterException {

		/*handle basic types*/
		if (expr.getType()==DeltaLexer.INT) {
			return DataType.INTEGER;
		}
		if (expr.getType()==DeltaLexer.STRING) {
			return DataType.STRING;
		}
		if (expr.getType()==DeltaLexer.FLOATING_POINT) {
			return DataType.FLOATING_POINT;
		}
		if (expr.getType()==DeltaLexer.FUNCTION_CALL) {
			return evalFunctionCall(expr);
		}

		/*if it is a variable, get the type from symbol table*/
		if (expr.getType()==DeltaLexer.ID) {
			checkVar(expr.getText());
			return symtab.get(expr.getText()).getDataType();
		}

		CommonTree left = (CommonTree)expr.getChild(0);
		DataType lefttype = evalExpr(left);

		/*void types can not be operated with anything*/
		if (lefttype==DataType.VOID) {
			throw new InterpreterException("type mismatch");
		}

		DataType righttype =  DataType.VOID;
		if (expr.getChildCount()==2) {
			CommonTree right = (CommonTree)expr.getChild(1);
			righttype =  evalExpr(right);
			if (righttype==DataType.VOID) {
				throw new InterpreterException("type mismatch");
			}
		}
		String operator = expr.getText();

		/*handle some special cases*/
		/*unary operator*/
		if (operator.equals("+") || operator.equals("-")) {
			if (expr.getChildCount()==1) {
				if (lefttype==DataType.STRING) {
					throw new InterpreterException("type mismatch");
				} else {
					return lefttype;
				}
			}
		}
		return TypeRule.getInstance().coerce(operator, lefttype, righttype);
	}

	void evalIf(CommonTree astmt) throws Exception {
		CommonTree expr = (CommonTree)astmt.getChild(0);
		DataType dt = evalExpr(expr);
		if (dt!=DataType.INTEGER) {
			throw new InterpreterException("Condition must be boolean/integer expression");
		}
		CommonTree ifpart = (CommonTree)astmt.getChild(1);
		evalStatementList(ifpart);
		if (astmt.getChildCount()==3) { /*else part*/
			CommonTree elsepart = (CommonTree)astmt.getChild(2);
			evalStatementList(elsepart);
		}
	}

	void evalWhile(CommonTree astmt) throws Exception {
		CommonTree expr = (CommonTree)astmt.getChild(0);
		DataType dt = evalExpr(expr);
		if (dt!=DataType.INTEGER) {
			throw new InterpreterException("Condition must be boolean/integer expression");
		}
		CommonTree whileblock = (CommonTree)astmt.getChild(1);
		evalStatementList(whileblock);
	}

	void evalDeclaration(CommonTree  decl) {
		CommonTree type = (CommonTree)decl.getChild(0);
		String typestr = type.getText();
		DataType datatype = DataType.VOID;
		ValueObject defaultval = null;
		if (typestr.equals("float")) {
			datatype = DataType.FLOATING_POINT;
		} else if (typestr.equals("string")) {
			datatype = DataType.STRING;
		} else if (typestr.equals("int")) {
			datatype = DataType.INTEGER;
		}

		for (int i = 1; i<decl.getChildCount(); i++) {
			CommonTree et = (CommonTree)decl.getChild(i);
			symtab.putVariable(et.getText(), datatype);
		}
	}

	void evalDeclarations(CommonTree  decl) {
		for (Object e: decl.getChildren()) {
			evalDeclaration((CommonTree)e);
		}
	}

	void evalAssignment(CommonTree astmt) throws Exception {
		CommonTree var = (CommonTree)astmt.getChild(0);
		CommonTree expr = (CommonTree)astmt.getChild(1);
		DataType left = DataType.VOID;
		DataType right = evalExpr(expr);
		if (var.getType()==DeltaLexer.ID) {
			checkVar(var.getText());
			left = symtab.get(var.getText()).getDataType();
		} else {
			throw new InterpreterException("Left side is not a variable");
		}
		TypeRule.getInstance().coerce("=", left, right);
	}


	void evalStatement(CommonTree expr) throws Exception{
		switch (expr.getType()) {
		case DeltaLexer.DECLARATIONS:
			evalDeclarations(expr);
			break;
		case DeltaLexer.FUNCTION_CALL:
			evalFunctionCall(expr);
			break;
		case DeltaLexer.ASSIGNMENT:
			evalAssignment(expr);
			break;
		case DeltaLexer.IF_STATEMENT:
			evalIf(expr);
			break;
		case DeltaLexer.WHILE_STATEMENT:
			evalWhile(expr);
			break;
		case DeltaLexer.STATEMENT_LIST:
			evalStatementList(expr);
			break;
		}
	}

	void evalStatementList(CommonTree exprlist) throws Exception{
		for (Object e: exprlist.getChildren())
			evalStatement((CommonTree)e);
	}

	void eval(CommonTree root) throws Exception {
		evalStatementList(root);
	}

	public DeltaTypeChecker(SymbolTable _symtab) {
		symtab = _symtab;
	}
}
